---------------------------------------------------------------------
-- Fichero:   traza.ads
-- Autor:     J.L.Villarroel, C.Sagues, J.D.Tardos
-- Version:   v2.2 14-3-2003
-- Proposito: Paquete para obtener trazas de tareas
---------------------------------------------------------------------
-- v2.0  27-3-2002  JDT
--    Ya no es un paquete generico
--    Eliminada funcion traza_completa
--    El numero maximo de anotaciones es 1000
-- v2.1 24-4-2002, JDT
--    Corregida la excepcion que aparecia si la traza esta vacia
-- v2.2 24-4-2002, JDT
--    Pasado a Ada.Real_Time y anadido parametro en comienzo_traza
---------------------------------------------------------------------

with Ada.Real_Time, text_io, ada.integer_text_io, ada.dynamic_priorities, system;
use  Ada.Real_Time, text_io, ada.integer_text_io, ada.dynamic_priorities;

package body traza is

  type clase_evento is (act,fin) ;
  package CEIO is new text_io.enumeration_io(clase_evento) ; use CEIO ;
  
  type indice_traza is range 1..1000 ; --numero_maximo_anotaciones ;
  type t_proceso is array (indice_traza) of positive ;
  type t_evento  is array (indice_traza) of clase_evento ;
  type t_tiempo  is array (indice_traza) of duration;

  package DIO is new text_io.fixed_io(duration) ; use DIO ;


  protected monitor is
    --pragma priority (system.priority'last) ;
    entry anota_evento (t: in positive; e: in clase_evento) ;
    procedure consulta_fin_traza (ft: out boolean) ;
    entry comienzo_traza(inicio: in time) ;
    procedure bloquea_traza ;
    procedure desbloquea_traza ;
    procedure escribe_traza(discretizacion: duration);                         
  private
    proceso: t_proceso ;
    evento: t_evento ;
    tiempo: t_tiempo ;
    max_tarea: positive:= 1;
    i: indice_traza;
    traza_completada: boolean := false ;
    tiempo_inicio_traza: time ;
    bloqueado: boolean := false ;
  end monitor ;
  
  protected body monitor is
  
    entry anota_evento (t: in positive; e: in clase_evento) when not bloqueado is
    begin
      if not traza_completada then
        if t > max_tarea then
          max_tarea:= t;
        end if;
        proceso(i) := t ;
        evento(i):= e ;
        tiempo(i):= To_Duration(clock - tiempo_inicio_traza) ;
        if i = indice_traza'last then 
          traza_completada:= true ;
        else
          i:= i + 1 ; 
        end if ;
      end if ;
    end anota_evento ;

    procedure consulta_fin_traza (ft: out boolean) is
    begin
      ft := traza_completada ;
    end consulta_fin_traza ;

    entry comienzo_traza(inicio: in time) when not bloqueado is
    begin
      if inicio = Time_First then
         tiempo_inicio_traza:= clock ;
      else
         tiempo_inicio_traza:= inicio;
      end if;
      i:= indice_traza'first ;
      traza_completada:= false ;
    end comienzo_traza ;
    
    procedure bloquea_traza is
    begin
      bloqueado := true ;
    end bloquea_traza ;

    procedure desbloquea_traza is
    begin
      bloqueado := false ;
    end desbloquea_traza ;

  procedure escribe_traza(discretizacion: duration) is                         
    type estado is (activa,espera) ;
    estado_tareas: array (1..max_tarea) of estado := (others => espera) ;
    escribe_tiempo: boolean := false ;
    k: indice_traza ;
    j: integer ;
    tiempo_evento: duration ;
    fin_representacion: boolean := false ;
  begin
    monitor.bloquea_traza ;
    put("Traza de ejecucion") ; new_line ;
    k:= 1 ; j:= 0 ;
    fin_representacion:= (k >= i); --v2.1
    while not fin_representacion loop
      escribe_tiempo:= false ;
      if j*discretizacion >= tiempo(k) then
        case evento(k) is
          when act => estado_tareas(proceso(k)) := activa ;
          when fin => estado_tareas(proceso(k)) := espera ;
        end case ;
        escribe_tiempo := true ;
        tiempo_evento:= tiempo(k) ;
        if k<i-1 then 
          k := k + 1 ; 
        else
          fin_representacion:= true ;
        end if ;
      else 
        j:= j + 1 ;
      end if ;
      for h in 1..max_tarea loop
        case estado_tareas(h) is
          when activa => put(h,2) ; put(" ") ;
          when espera => put ("   ") ;
        end case ;
      end loop ;
      if escribe_tiempo then put (tiempo_evento) ; end if ;
      new_line ;
    end loop ;
    put("Fin de la Traza") ; new_line ;
    monitor.desbloquea_traza ;
  end escribe_traza ;

end monitor ;

-- Funciones publicas del paquete Traza ---------------------------------------

  procedure comienzo_traza(inicio: in time:= Time_First) is
  current_prio : System.Any_Priority ; 
  begin
    current_prio := get_priority ;
    set_priority (20) ; --maxima prioridad
    monitor.comienzo_traza(inicio);
    set_priority (current_prio) ; 
  end comienzo_traza ;
  
  procedure ejecuto (tarea: in positive) is
  current_prio : System.Any_Priority ; 
  begin
    current_prio := get_priority ;
    set_priority (20) ;
    monitor.anota_evento (tarea,act) ;
    set_priority (current_prio) ;  
  end ejecuto ;
      
  
  procedure termino (tarea: in positive)  is
   current_prio : System.Any_Priority ; 
  begin
    current_prio := get_priority ;
    set_priority (20) ;--maxima prioridad
    monitor.anota_evento (tarea,fin) ;
    set_priority (current_prio) ;  
  end termino ;
  
  function traza_completa return boolean is
    traza_completada: boolean ;
    current_prio : System.Any_Priority ; 
  begin
    current_prio := get_priority ;
    set_priority (20) ;--maxima prioridad
    monitor.consulta_fin_traza (traza_completada) ;
    set_priority (current_prio) ; 
    return traza_completada ;
  end ;
  
  procedure escribe_traza(discretizacion: duration:= 0.5) is                         
  current_prio : System.Any_Priority ; 
  begin
    current_prio := get_priority ;
    set_priority (20) ;--maxima prioridad
    monitor.escribe_traza(discretizacion) ;
    set_priority (current_prio) ; 
  end escribe_traza ;
  
end traza ;
