#include "yenta_socket.h"
#include "i82365.h"


/*
* Generate easy-to-use ways of reading a cardbus sockets
* regular memory space ("cb_xxx"), configuration space
* ("config_xxx") and compatibility space ("exca_xxxx")
*/
static inline u32 cb_readl(struct yenta_socket *socket, unsigned reg)
{
    u32 val = readl(socket->base + reg);
    return val;
}

static inline void cb_writel(struct yenta_socket *socket, unsigned reg, u32 val)
{
    writel(val, socket->base + reg);
    readl(socket->base + reg); /* avoid problems with PCI write posting */
}

static inline u8 config_readb(struct yenta_socket *socket, unsigned offset)
{
    u8 val;
    pci_read_config_byte(socket->dev, offset, &val);
    return val;
}

static inline u16 config_readw(struct yenta_socket *socket, unsigned offset)
{
    u16 val;
    pci_read_config_word(socket->dev, offset, &val);
    return val;
}

static inline u32 config_readl(struct yenta_socket *socket, unsigned offset)
{
    u32 val;
    pci_read_config_dword(socket->dev, offset, &val);
    return val;
}

static inline void config_writeb(struct yenta_socket *socket, unsigned offset, u8 val)
{
    pci_write_config_byte(socket->dev, offset, val);
}

static inline void config_writew(struct yenta_socket *socket, unsigned offset, u16 val)
{
    pci_write_config_word(socket->dev, offset, val);
}

static inline void config_writel(struct yenta_socket *socket, unsigned offset, u32 val)
{
    pci_write_config_dword(socket->dev, offset, val);
}

static inline u8 exca_readb(struct yenta_socket *socket, unsigned reg)
{
    u8 val = readb(socket->base + 0x800 + reg);
    return val;
}

static inline u8 exca_readw(struct yenta_socket *socket, unsigned reg)
{
    u16 val;
    val = readb(socket->base + 0x800 + reg);
    val |= readb(socket->base + 0x800 + reg + 1) << 8;
    return val;
}

static inline void exca_writeb(struct yenta_socket *socket, unsigned reg, u8 val)
{
    writeb(val, socket->base + 0x800 + reg);
    readb(socket->base + 0x800 + reg); /* PCI write posting... */
}

// static void exca_writew(struct yenta_socket *socket, unsigned reg, u16 val)
// {
//     writeb(val, socket->base + 0x800 + reg);
//     writeb(val >> 8, socket->base + 0x800 + reg + 1);
//
//     /* PCI write posting... */
//     readb(socket->base + 0x800 + reg);
//     readb(socket->base + 0x800 + reg + 1);
// }

/*
static ssize_t show_yenta_registers(struct pci_device *dev, struct yenta_socket *socket)
{
    char buf[256];
    int offset = 0, i;

    offset = snprintf(buf, PAGE_SIZE, "CB registers:");
    for (i = 0; i < 0x24; i += 4) {
        unsigned val;
        if (!(i & 15))
            offset += snprintf(buf + offset, PAGE_SIZE - offset, "\n%02x:", i);
        val = cb_readl(socket, i);
        offset += snprintf(buf + offset, PAGE_SIZE - offset, " %08x", val);
    }

    offset += snprintf(buf + offset, PAGE_SIZE - offset, "\n\nExCA registers:");
    for (i = 0; i < 0x45; i++) {
        unsigned char val;
        if (!(i & 7)) {
            if (i & 8) {
                memcpy(buf + offset, " -", 2);
                offset += 2;
            } else
                offset += snprintf(buf + offset, PAGE_SIZE - offset, "\n%02x:", i);
        }
        val = exca_readb(socket, i);
        offset += snprintf(buf + offset, PAGE_SIZE - offset, " %02x", val);
    }
    buf[offset++] = '\n';
    return offset;
}*/

/*
* Ugh, mixed-mode cardbus and 16-bit pccard state: things depend
* on what kind of card is inserted..
*/
static int yenta_get_status(struct yenta_socket *socket, unsigned int *value)
{
    unsigned int val;
    u32 state = cb_readl(socket, CB_SOCKET_STATE);

    val  = (state & CB_3VCARD) ? SS_3VCARD : 0;
    val |= (state & CB_XVCARD) ? SS_XVCARD : 0;
    val |= (state & (CB_5VCARD | CB_3VCARD | CB_XVCARD | CB_YVCARD)) ? 0 : SS_PENDING;
    val |= (state & (CB_CDETECT1 | CB_CDETECT2)) ? SS_PENDING : 0;


    if (state & CB_CBCARD) {
        printf("getstatus: cardbus card %X\n", state);
        val |= SS_CARDBUS;
        val |= (state & CB_CARDSTS) ? SS_STSCHG : 0;
        val |= (state & (CB_CDETECT1 | CB_CDETECT2)) ? 0 : SS_DETECT;
        val |= (state & CB_PWRCYCLE) ? SS_POWERON | SS_READY : 0;
    } else if (state & CB_16BITCARD) {
        printf("getstatus: 16bit card\n");
        u8 status = exca_readb(socket, I365_STATUS);
        val |= ((status & I365_CS_DETECT) == I365_CS_DETECT) ? SS_DETECT : 0;
        if (exca_readb(socket, I365_INTCTL) & I365_PC_IOCARD) {
            val |= (status & I365_CS_STSCHG) ? 0 : SS_STSCHG;
        } else {
            val |= (status & I365_CS_BVD1) ? 0 : SS_BATDEAD;
            val |= (status & I365_CS_BVD2) ? 0 : SS_BATWARN;
        }
        val |= (status & I365_CS_WRPROT) ? SS_WRPROT : 0;
        val |= (status & I365_CS_READY) ? SS_READY : 0;
        val |= (status & I365_CS_POWERON) ? SS_POWERON : 0;
    }

    *value = val;
    return 0;
}

static void yenta_set_power(struct yenta_socket *socket, socket_state_t *state)
{
    u32 reg = 0;    /* CB_SC_STPCLK? */

    switch (state->Vcc) {
        case 33: reg = CB_SC_VCC_3V; break;
        case 50: reg = CB_SC_VCC_5V; break;
        default: reg = 0; break;
    }
    switch (state->Vpp) {
        case 33:  reg |= CB_SC_VPP_3V; break;
        case 50:  reg |= CB_SC_VPP_5V; break;
        case 120: reg |= CB_SC_VPP_12V; break;
    }

    if (reg != cb_readl(socket, CB_SOCKET_CONTROL))
        cb_writel(socket, CB_SOCKET_CONTROL, reg);
}

static int yenta_set_socket(struct yenta_socket *socket, socket_state_t *state)
{
    u16 bridge;
    printf("yenta_set_socket state->Vcc == %d\n",state->Vcc);

    /* if powering down: do it immediately */
    if (state->Vcc == 0)
        yenta_set_power(socket, state);

    state->io_irq = socket->cb_irq;
    socket->io_irq = state->io_irq;

    bridge = config_readw(socket, CB_BRIDGE_CONTROL) & ~(CB_BRIDGE_CRST | CB_BRIDGE_INTR);
    if (cb_readl(socket, CB_SOCKET_STATE) & CB_CBCARD) {
        u8 intr;
        bridge |= (state->flags & SS_RESET) ? CB_BRIDGE_CRST : 0;

        /* ISA interrupt control? */
        intr = exca_readb(socket, I365_INTCTL);
        intr = (intr & ~0xf);
        if (!socket->cb_irq) {
            intr |= state->io_irq;
            bridge |= CB_BRIDGE_INTR;
        }
        exca_writeb(socket, I365_INTCTL, intr);
    }  else {
        u8 reg;

        reg = exca_readb(socket, I365_INTCTL) & (I365_RING_ENA | I365_INTR_ENA);
        reg |= (state->flags & SS_RESET) ? 0 : I365_PC_RESET;
        reg |= (state->flags & SS_IOCARD) ? I365_PC_IOCARD : 0;
        if (state->io_irq != socket->cb_irq) {
            reg |= state->io_irq;
            bridge |= CB_BRIDGE_INTR;
        }
        exca_writeb(socket, I365_INTCTL, reg);

        reg = exca_readb(socket, I365_POWER) & (I365_VCC_MASK|I365_VPP1_MASK);
        reg |= I365_PWR_NORESET;
        if (state->flags & SS_PWR_AUTO) reg |= I365_PWR_AUTO;
        if (state->flags & SS_OUTPUT_ENA) reg |= I365_PWR_OUT;
        if (exca_readb(socket, I365_POWER) != reg)
            exca_writeb(socket, I365_POWER, reg);

        /* CSC interrupt: no ISA irq for CSC */
        reg = I365_CSC_DETECT;
        if (state->flags & SS_IOCARD) {
            if (state->csc_mask & SS_STSCHG) reg |= I365_CSC_STSCHG;
        } else {
            if (state->csc_mask & SS_BATDEAD) reg |= I365_CSC_BVD1;
            if (state->csc_mask & SS_BATWARN) reg |= I365_CSC_BVD2;
            if (state->csc_mask & SS_READY) reg |= I365_CSC_READY;
        }
        exca_writeb(socket, I365_CSCINT, reg);
        exca_readb(socket, I365_CSC);
//         if(sock->zoom_video)
//             sock->zoom_video(sock, state->flags & SS_ZVCARD);
    }
    config_writew(socket, CB_BRIDGE_CONTROL, bridge);
    /* Socket event mask: get card insert/remove events.. */
    cb_writel(socket, CB_SOCKET_EVENT, -1);
    cb_writel(socket, CB_SOCKET_MASK, CB_CDMASK);

    /* if powering up: do it as the last step when the socket is configured */
    if (state->Vcc != 0)
        yenta_set_power(socket, state);
    return 0;
}

static int yenta_interrupt(void* dev_id, intr_t irq)
{
    unsigned int events;
    struct yenta_socket *socket = (struct yenta_socket *) dev_id;
    u8 csc;
    u32 cb_event;

    //printk("irq!\n");
    /* Clear interrupt status for the event */
    cb_event = cb_readl(socket, CB_SOCKET_EVENT);
    cb_writel(socket, CB_SOCKET_EVENT, cb_event);

    csc = exca_readb(socket, I365_CSC);

    if (!(cb_event || csc))
        return POSIX_INTR_HANDLED_NOTIFY;

    events = (cb_event & (CB_CD1EVENT | CB_CD2EVENT)) ? SS_DETECT : 0 ;
    events |= (csc & I365_CSC_DETECT) ? SS_DETECT : 0;
    if (exca_readb(socket, I365_INTCTL) & I365_PC_IOCARD) {
        events |= (csc & I365_CSC_STSCHG) ? SS_STSCHG : 0;
    } else {
        events |= (csc & I365_CSC_BVD1) ? SS_BATDEAD : 0;
        events |= (csc & I365_CSC_BVD2) ? SS_BATWARN : 0;
        events |= (csc & I365_CSC_READY) ? SS_READY : 0;
    }

//  if (events)
//      pcmcia_parse_events(&socket->socket, events);

    return POSIX_INTR_HANDLED_NOTIFY;
}

/* redoes voltage interrogation if required */
static void yenta_interrogate(struct yenta_socket *socket)
{
    u32 state;

    state = cb_readl(socket, CB_SOCKET_STATE);
    if (!(state & (CB_5VCARD | CB_3VCARD | CB_XVCARD | CB_YVCARD)) ||
          (state & (CB_CDETECT1 | CB_CDETECT2 | CB_NOTACARD | CB_BADVCCREQ)) ||
          ((state & (CB_16BITCARD | CB_CBCARD)) == (CB_16BITCARD | CB_CBCARD)))
        cb_writel(socket, CB_SOCKET_FORCE, CB_CVSTEST);
}

socket_state_t dead_socket = {
    .csc_mask   = SS_DETECT,
};

static void yenta_clear_maps(struct yenta_socket *socket)
{
//     int i;
//     struct resource res = { .start = 0, .end = 0x0fff };
//     pccard_io_map io = { 0, 0, 0, 0, 1 };
//     pccard_mem_map mem = { .res = &res, };

    yenta_set_socket(socket, &dead_socket);
//     for (i = 0; i < 2; i++) {
//         io.map = i;
//         yenta_set_io_map(&socket->socket, &io);
//     }
//     for (i = 0; i < 5; i++) {
//         mem.map = i;
//         yenta_set_mem_map(&socket->socket, &mem);
//     }
}

//
// /* Called at resume and initialization events */
static int yenta_sock_init(struct yenta_socket *socket)
{
    exca_writeb(socket, I365_GBLCTL, 0x00);
    exca_writeb(socket, I365_GENCTL, 0x00);

    /* Redo card voltage interrogation */
    yenta_interrogate(socket);

    yenta_clear_maps(socket);

//  if (socket->type && socket->type->sock_init)
//      socket->type->sock_init(socket);

    /* Re-enable CSC interrupts */
    cb_writel(socket, CB_SOCKET_MASK, CB_CDMASK);

    return 0;
}

/*
* Allocate the bridge mappings for the device..
*/
static void yenta_allocate_resources(struct yenta_socket *socket)
{
    struct pci_bus_region region;
    struct pci_device *bridge = socket->dev;

//     region.start = 0x1000;
//     region.end = 0x10FF;

    region.start = 0x1400;
    region.end = 0x14FF;
    printk("  IO window: %08lx-%08lx\n", region.start, region.end);
    pci_write_config_dword(bridge, PCI_CB_IO_BASE_0, region.start);
    pci_write_config_dword(bridge, PCI_CB_IO_LIMIT_0, region.end);

//     region.start = 0x1400;
//     region.end = 0x14FF;
    region.start = 0x1800;
    region.end = 0x18FF;
    printk("  IO window: %08lx-%08lx\n", region.start, region.end);
    pci_write_config_dword(bridge, PCI_CB_IO_BASE_1, region.start);
    pci_write_config_dword(bridge, PCI_CB_IO_LIMIT_1, region.end);

//     region.start = 0x10000000;
//     region.end = 0x13FFF000;
    region.start = 0x20000000;
    region.end = 0x21FFF000;
    printk("  PREFETCH window: %08lx-%08lx\n", region.start, region.end);
    pci_write_config_dword(bridge, PCI_CB_MEMORY_BASE_0, region.start);
    pci_write_config_dword(bridge, PCI_CB_MEMORY_LIMIT_0, region.end);

    region.start = 0x22000000;
    region.end = 0x23FFF000;
    printk("  MEM window: %08lx-%08lx\n",region.start, region.end);
    pci_write_config_dword(bridge, PCI_CB_MEMORY_BASE_1, region.start);
    pci_write_config_dword(bridge, PCI_CB_MEMORY_LIMIT_1, region.end);
}

// #include "ricoh.h"
#include "ti113x.h"

/*
* Initialize the standard cardbus registers
*/
static void yenta_config_init(struct yenta_socket *socket)
{
    u16 bridge;
    struct pci_device *dev = socket->dev;

    config_writel(socket, CB_LEGACY_MODE_BASE, 0);
    config_writel(socket, PCI_BASE_ADDRESS_0, (u32)socket->base);
    config_writew(socket, PCI_COMMAND,
                  PCI_COMMAND_IO |
                          PCI_COMMAND_MEMORY |
                          PCI_COMMAND_MASTER |
                          PCI_COMMAND_WAIT);

    /* MAGIC NUMBERS! Fixme */
    config_writeb(socket, PCI_CACHE_LINE_SIZE, 0x04);
    config_writeb(socket, PCI_LATENCY_TIMER, 168);
    config_writel(socket, PCI_PRIMARY_BUS,
                  (176 << 24) |              /* sec. latency timer */
                  (dev->subordinate << 16) | /* subordinate bus */
                  (dev->secondary << 8) |  /* secondary bus */
                   dev->primary);        /* primary bus */

    /*
    * Set up the bridging state:
    *  - enable write posting.
    *  - memory window 0 prefetchable, window 1 non-prefetchable
    *  - PCI interrupts enabled if a PCI interrupt exists..
    */
    bridge = config_readw(socket, CB_BRIDGE_CONTROL);
    bridge &= ~(CB_BRIDGE_CRST | CB_BRIDGE_PREFETCH1 | CB_BRIDGE_ISAEN | CB_BRIDGE_VGAEN);
    bridge |= CB_BRIDGE_PREFETCH0 | CB_BRIDGE_POSTEN;
    config_writew(socket, CB_BRIDGE_CONTROL, bridge);
}

static int socket_reset(struct yenta_socket *socket, socket_state_t *state)
{
    int status, i;

    printf("reset\n");

    state->flags |= SS_OUTPUT_ENA | SS_RESET;
    yenta_set_socket(socket, state);
    udelay((long)10);

    state->flags &= ~SS_RESET;
    yenta_set_socket(socket, state);

    msleep(10 * 10);
    for (i = 0; i < 30; i++) {
        yenta_get_status(socket, &status);

        if (!(status & SS_DETECT))
            return CS_NO_CARD;

        if (status & SS_READY)
            return CS_SUCCESS;

        msleep(10 * 10);
    }
    printf("ERROR in reset\n");
    return CS_GENERAL_FAILURE;
}

static int socket_setup(struct yenta_socket *socket, int initial_delay)
{
    int i;
    unsigned int status;
    socket_state_t state;

    msleep(initial_delay * 10);

    for (i = 0; i < 100; i++) {
        yenta_get_status(socket, &status);
        if (!(status & SS_DETECT))
            return CS_NO_CARD;

        if (!(status & SS_PENDING))
            break;

        msleep(100);
    }

    if (status & SS_PENDING) {
        printf("ERROR socket_setup SS_PENDING\n");
        return -1;
    }
	//state.Vcc = state.Vpp = 33; //DANI

    if (status & SS_3VCARD) {
        state.Vcc = state.Vpp = 33;
        printf("socket is SS_3VCARD\n");
    } else if (!(status & SS_XVCARD)) {
        state.Vcc = state.Vpp = 50;
    } else {
        printf("unsupported voltage key.\n");
        return -1;
    }

//     if (skt->power_hook)
//         skt->power_hook(skt, HOOK_POWER_PRE);

    state.flags = 0;
    yenta_set_socket(socket, &state);

    /*
    * Wait "vcc_settle" for the supply to stabilise.
    */
    msleep(40 * 10);

    yenta_get_status(socket, &status);
    if (!(status & SS_POWERON)) {
        printf("unable to apply power.\n");
        return -1;
    }

    status = socket_reset(socket, &state);

//     if (skt->power_hook)
//         skt->power_hook(skt, HOOK_POWER_POST);
    return status;
}


/*
* Initialize a cardbus controller. Make sure we have a usable
* interrupt, and that we can map the cardbus area. Fill in the
* socket information structure..
*/
int yenta_probe (struct pci_device *dev, const struct pci_device_id *id)
{
    struct yenta_socket *socket;
    int ret = 0;
    socket_state_t state;
    unsigned int pcmcia_socket_status;

    socket = kzalloc(sizeof(struct yenta_socket), GFP_KERNEL);
    if (!socket) {
        printf("error kzalloc\n");
        return -ENOMEM;
    }

    /* prepare struct yenta_socket */
    socket->dev = dev;
    socket->base = (void *)dev->membase;

    yenta_config_init(socket);

    /* Disable all events */
    cb_writel(socket, CB_SOCKET_MASK, 0x0);

    /* Set up the bridge regions.. */
    yenta_allocate_resources(socket);

    socket->cb_irq = dev->irq;

    /* Do we have special options for the device? */
//     ricoh_override(socket);
ti_init(socket);
    ti113x_override(socket);


    /* We must finish initialization here */
    posix_intr_associate(socket->cb_irq, yenta_interrupt, socket, sizeof(struct yenta_socket));
    posix_intr_unlock(socket->cb_irq);

    /* Figure out what the dang thing can do for the PCMCIA layer... */
    yenta_interrogate(socket);

    /* get_socket_capabilities da esto: */
    socket->socket.irq_mask = 0x0007;
    socket->cb_irq = 9; //WAS 5
    //------------------------------------------------------------------//
    yenta_sock_init(socket);

    state.flags = 0;
    state.Vcc = 0;
    state.Vpp = 0;
    yenta_set_socket(socket, &state);
//  printf("Voy a detectar\n");
    // socket_detect_change
    yenta_get_status(socket, &pcmcia_socket_status);

// while (!(pcmcia_socket_status & SS_DETECT)){yenta_get_status(socket, &pcmcia_socket_status);
// sleep(1);printf("Try cardbus card... socket_insert\n");}
    if (pcmcia_socket_status & SS_DETECT) {
        // socket_insert
        printf("Detected cardbus card... socket_insert\n");
        ret = socket_setup(socket, 10);
        if (ret == CS_SUCCESS) {
            printf("pccard: CardBus card inserted into slot\n");
        } else {
            printf("socketsetup failed\n");
            return -1;
        }
        // cb_alloc(skt);
        // send_event(skt, CS_EVENT_CARD_INSERTION, CS_EVENT_PRI_LOW);
    }
    return 0;
}
